package LLI::Cell;

# This package will be used to define cells on the world.
#
# You can initialize it with a lat/lon position object and get the cell values
# that the lat/lon belongs to.
#
# $Id: Cell.pm,v 1.1 2003/03/24 01:05:11 mankins Exp $

use strict;

use LLI::Position;

my $VERSION = 1.00;

sub new 
  {
    my $proto = shift;
    my $class = ref($proto) || $proto;

    my ($position) = (@_);
    
    my $self = {};
    bless ($self, $class);

    # we can construct this with a position object, or a latlon hash, cell nameor undef

    if (ref $position =~ /Position/)
      {
	$self->{'position'} = $position; # a position object
      }
    elsif (ref $position eq "HASH")
      {
	my ($lat,$lon) = ($position->{'lat'},$position->{'lon'});
	next unless (defined $lat && defined $lon);

	# populate the position object
	my $pos = new LLI::Position;
	$pos->units_latlon("sexagesimal");
	$pos->lat($lat);
	$pos->lon($lon);

	$self->{'position'} = $pos;
      }
    elsif ($position =~ /^c\:/)
      {
	# it's a cell name
	# c:10Dlat,10Dlon:1Dlat,1Dlon:Mlat,Mlon:Slat,Slon
	# c:-10,150:2,9:50,1:23,59 (cell)

	my ($lat,$lon) = $self->cell_to_latlon($position);

	my $pos = new LLI::Position;
	$pos->units_latlon("sexagesimal");
	$pos->lat($lat);
	$pos->lon($lon);

	$self->{'position'} = $pos;
      }
    else
      {
	# nothing/invalid position given to us

	my $pos = new LLI::Position;
	$self->{'position'} = $pos;
      }

    $self->calculate_cells();

    return $self;
  }

sub position
  {
    # the postion object
    my $self = shift;

    my ($position) = @_;

    if ($position)
      {
	$self->{'position'} = $position; # a LLI::Position object
      }

    return $self->{'position'};
  }

sub get_cell_name
  {
    # returns the simplified name for the cell
    my $self = shift;
    
    # maybe we should make sure these are the correct # of digits? (again?)
    
    $self->calculate_cells(); # we shouldn't need to do this here...
        
    my $cell_name;
    $cell_name = $self->degrees_10("lat").",".$self->degrees_10("lon").":"; 
    $cell_name .= $self->degrees_1("lat").",".$self->degrees_1("lon").":"; 
    $cell_name .= $self->minutes("lat").",".$self->minutes("lon").":"; 
    $cell_name .= $self->seconds("lat").",".$self->seconds("lon"); 

    return $cell_name;
  }

sub get_latlon_parts
  {
    # split up the cell name into its parts for direct access.

    my $self = shift;

    my %pos;

    $pos{"10_lat"} = $self->degrees_10("lat");
    $pos{"10_lon"} = $self->degrees_10("lon");

    $pos{"1_lat"} = $self->degrees_1("lat");
    $pos{"1_lon"} = $self->degrees_1("lon");

    $pos{"min_lat"} = $self->minutes("lat");
    $pos{"min_lon"} = $self->minutes("lon");

    $pos{"sec_lat"} = $self->seconds("lat");
    $pos{"sec_lon"} = $self->seconds("lon");

    return \%pos;
  }

sub cell_to_latlon
  {
    # convert from cell space to latlons [in 60s]
    my $self = shift;

    my ($cell_name) = @_;

    # remove any namespace junk
    $cell_name =~ s/^c\://i;

    my (@parts) = split(':',$cell_name);

    my ($lat,$lon,$count);

    foreach my $part (@parts)
      {
	my ($latitude,$longitude) = split(',',$part);

	if ($count != 1)
	  {
	    $lat .= $latitude . ":";
	    $lon .= $longitude . ":";
	  }
	else
	  {
	    # combine the 10's and 1's 
	    chop($lat);
	    chop($lon);

	    my $sign_lat = ($lat > 0) ? 1 : -1;
	    my $sign_lon = ($lon > 0) ? 1 : -1;

	    $lat = abs($lat) + abs($latitude);
	    $lon = abs($lon) + abs($longitude);

	    $lat *= $sign_lat;
	    $lon *= $sign_lon;

	    $lat .= ":";
	    $lon .= ":";
	  }

	$count++;
      }

    chop ($lat);
    chop ($lon); # remove trailing :
    
    return ($lat,$lon);
  }

sub calculate_cells
  {
    # figures out what cells we're in. 

    my $self = shift;

    # Set to 60s mode
    $self->{'position'}->units_latlon("sexagesimal");

    my ($lat) = $self->{'position'}->lat();
    my ($lon) = $self->{'position'}->lon();

    # +DDD:MM:SS

  DO_LAT:
    {
      my ($units,$minutes,$seconds) = split(':',$lat);

      # get the tens 
      my ($tens) = int($units/10) . "0";
      $tens = "0" if (int($tens) == 0);

      $self->degrees_10("lat",$tens);
      $self->degrees_1("lat",abs($units) - abs($tens)); 
      $self->minutes("lat",$minutes);
      $self->seconds("lat",$seconds);
    }

  DO_LON:
    {
      my ($units,$minutes,$seconds) = split(':',$lon);

      # get the tens 
      my ($tens) = int($units/10) . "0";
      $tens = "0" if (int($tens) == 0);

      $self->degrees_10("lon",$tens);
      $self->degrees_1("lon",abs($units) - abs($tens)); 
      $self->minutes("lon",$minutes);
      $self->seconds("lon",$seconds);
    }
  }

sub degrees_10
  {
    my $self = shift;

    my ($type,$value) = @_;

    if (defined $value)
      {
	$self->{$type}->{'10_degree'} = $value;
      }
    
    return $self->{$type}->{'10_degree'};
  }

sub degrees_1
  {
    my $self = shift;

    my ($type,$value) = @_;

    if (defined $value)
      {
	$self->{$type}->{'1_degree'} = $value;
      }
    
    return $self->{$type}->{'1_degree'};
  }

sub minutes
  {
    my $self = shift;

    my ($type,$value) = @_;

    if (defined $value)
      {
	$self->{$type}->{'minutes'} = $value;
      }
    
    return $self->{$type}->{'minutes'};
  }

sub seconds
  {
    my $self = shift;

    my ($type,$value) = @_;

    if (defined $value)
      {
	$self->{$type}->{'seconds'} = $value;
      }
    
    return $self->{$type}->{'seconds'};
  }

1;
