#!/usr/bin/perl

# Search bot for the LLI component (Location Linked Information)
# $Id: lli-searcher.pl,v 1.3 2003/04/01 22:40:31 mankins Exp $
#
# (c) 2003 Matt Mankins - http://www.media.mit.edu/~mankins/lli/
#
# This code is awoken by a signal which causes it to look at the 
# search queue in the database and then perform the searches that
# reside there.

package LLI::Searcher;

use strict;

use DBI;
use XML::Simple;
use Term::ANSIColor;
use Data::Dumper;
use Jabber::Connection;
use Jabber::NodeFactory;

use LLI::NodeStore;
use LLI::RouteTable;

# Load Config File
my $cfg = XMLin("lli.xml") || die "Can't load config file: $!\n";
$cfg = XMLin("lli-debug.xml") if (@ARGV); # allow us to have another server with same code base

# User Defines
my $sleep_time = $cfg->{'searcher'}->{"sleep_time"} || 60;

# Setup Database Handle, Store connection, Component Connection
my $dbh = DBI->connect($cfg->{'database'}->{'dbh'},$cfg->{'database'}->{'user'},$cfg->{'database'}->{'password'});
my $store = new LLI::NodeStore ( { "dbh" => $dbh } );

# Component connection

my $c = new Jabber::Connection(
			       server    => 
			       $cfg->{'component'}->{'searcher'}->{'host'}.":".
			       $cfg->{'component'}->{'searcher'}->{'port'},
			       
			       localname => $cfg->{'component'}->{'searcher'}->{'name'},
			       ns        => 'jabber:component:accept',
			       debug     => $cfg->{'debug'}->{'level'},
			      );

$c->register_handler('iq', \&handle_iq);

my $route = new LLI::RouteTable (
				 { 'dbh' => $dbh,
				   'primer_host' => $cfg->{"routes"}->{"primer"},
				   'jabber_from' => $cfg->{"component"}->{"publichost"}
				 }
				);

DEBUGGING:
{
  open (DEBUG, ">>$cfg->{'debug'}->{'file'}");
  select DEBUG;
  $| = 1; # make debug unbuffered
  select STDOUT;
  sub debug;
}

SIGNAL_HANDLERS:
{
  # Setup Signal Handlers

  $SIG{KILL} = $SIG{TERM} = $SIG{INT} = \&cleanup;
  $SIG{HUP} = $SIG{USR1} = \&start_searches;

  debug "signal handlers setup";
}

INITIALIZE_COMPONENT_CHANNEL:
{
  # our component will communicate with us through signal they send to us, so they need
  # to know what our process id is.
  
  my $pid_file =  $cfg->{"searcher"}->{"pidfile"} || "searcher.pid";

  open (PID, ">$pid_file") || warn ("can't open pid file $pid_file $!\n");
  print PID $$."\n";
  close PID;

  debug "pid $$";
}

CONNECT_TO_JABBER_SERVER:
{
  # We use the Jabber::Connection module for this component 
  # because it has a looser event model and doesn't make as many assumptions

  debug "connecting...pw:".$cfg->{'component'}->{'searcher'}->{'secret'};

  die "can't connect to jabber server: ".$c->lastError unless $c->connect();

  $c->auth($cfg->{'component'}->{'searcher'}->{'secret'});

  debug "connected to server";
}

MAIN_LOOP:
{
  debug "starting main loop...";

  while (1)
    {
      # Clear our any existing searches in our queue
      &start_searches(1);

      # Then we just wait.
      sleep $sleep_time;

      # Even though we weren't given a signal, we'll check the search queue
      # anyway.
    }
}

sub start_searches
  {
    debug ("starting search",time,@_);

    my $pending = $store->pending_searches();

    debug $pending;

    foreach my $pending_search (keys %{ $pending })
      {
	debug "searching for $pending_search";
	
	my ($is_valid);
	($is_valid,$pending->{$pending_search}) = is_valid_search($pending->{$pending_search});

	my ($local_results);

	if ($is_valid)
	  {
	    debug ("valid search",$pending->{$pending_search});
	    
	    # look locally to see if we have any nodes that match
	    $local_results = &local_search($pending->{$pending_search});

	    # if we have matches, relay them to the return node
	    if (scalar (keys %{ $local_results }) >= 1)
	      {
		debug (scalar(keys %{ $local_results }), " results found, relaying");
		relay_results($local_results,$pending_search);
	      }
	    else
	      {
		debug "no local results found, will still try to relay to remote hosts.";
	      }

	    my $success = &remote_search($pending->{$pending_search},$pending_search);
	    unless ($success)
	      {
		debug "remote search was not successful";
	      }
	  }
	else
	  {
	    debug "search not valid :(";
	  }
	
	$store->mark_search_processed($pending_search);
	#delete_search($pending_search); # we want this to be done elsewhere to save the data.
      }
  }

sub relay_results
  {
    # we'll publish this data to the return node
    # this will always be local data, published to a local node.

    my ($results,$searchid) = @_;

    # 'b2cb0ae4462f8dc61562f451843e01e9b04a8af0' => {
    # 'resource' => undef,
    #'cost' => undef,
    #'cache' => undef,
    #'nuggetid' => 'b2cb0ae4462f8dc61562f451843e01e9b04a8af0',
    #'description' => undef,
    #'rowid' => '41',
    #'expiration' => '2003-06-17 15:21:44',
    #'item_name' => '1048101704',
    #'lat' => '33.3330560',
    #'value' => undef,
    #'source' => undef,
    #'nodeid' => '17',
    #'name' => 'p:tester5@jabber.media.mit.edu',
    #'alt' => undef,
    #'url' => undef,
    #'last_modified' => '20030319142144',
    #'lon' => '44.2400000',
    #'owner' => 'tester5@jabber.media.mit.edu',
    #'type' => undef
    #}

    debug ("relaying for $searchid and ->",$results);

    # Lookup where to send the data: what node?
    my ($destination_info) = $store->get_search_info($searchid);
    my ($node_name) = $destination_info->{$searchid}->{"node_name"};

    debug ("destination info:",$destination_info,$node_name);

    # now for each search result (there could be many), relay
    foreach my $nuggetid (keys %{ $results })
      {
	debug "doing $nuggetid";

	my $content = {
		       'node' => $node_name,
		       'item' => [
				  {
				   'id' => $nuggetid,
				   
				   'position' =>
				   {
				    'xmlns' => 'lli-position',
				
				    'lat' => [$results->{$nuggetid}->{"lat"}],
				    'lon' => [$results->{$nuggetid}->{"lon"}],
				    'alt' => [$results->{$nuggetid}->{"alt"}],
				   },
				   'link' =>
				   {
				    'xmlns' => 'lli-link',
				    'url'   => [$results->{$nuggetid}->{"url"}],
				    'cost'  => [$results->{$nuggetid}->{"cost"}],	
				   },
				   'data' =>
				   {
				    'xmlns'       => 'lli-data',
				    'description' => [$results->{$nuggetid}->{"description"}],
				    'type'        => [$results->{$nuggetid}->{"type"}],
				   },
				   'time' =>
				   {
				    'xmlns' => 'lli-time',
				    'expiration' => [$results->{$nuggetid}->{"expiration"}],
				   },
				  },
				 ]
		      };

	# we relay this data from ourselves to our "parent" component.
	my $component = $cfg->{"component"}->{"publichost"};
	debug ("publishing to $component at this node $node_name",$content);
	
	my $response = publish_node($component, $node_name, $content);
	# do something with the response--check for error?
      }
  }

sub publish_node
  {
    my ($component, $node_name, $content) = @_;

    debug "publishing to $component, $node_name and $content";

    my $pubsub = {
                  'iq'  => { 'type' => 'set',
                             'from' => $cfg->{'component'}->{'searcher'}->{'publichost'},
                             'to'   => $component,
			     'id'   => time.$$,
                             'pubsub' => { 'xmlns' => 'http://jabber.org/protocol/pubsub',
                                           'publish' => $content
                                         }
                           }
                 };

    my ($xml) = create_xml($pubsub);

    debug "publishing: $xml";

    my $nf = new Jabber::NodeFactory(fromstr => 1);
    my $node = $nf->newNodeFromStr($xml);

    my ($response) = $c->ask($node);
    debug "response: $response";
    return unless $response;
    $response = $response->toStr();

    debug ("send got back",$response);
    return $response;
  }

sub local_search
  {
    my ($search) = @_;

    debug "doing local search";

    my $results = $store->do_search($search);
    debug $results;
    debug "results!";

    return $results;
  }

sub remote_search
  {
    my ($search,$searchid) = @_;

    debug ("doing remote search",$search);

    my ($lat,$lon) = ($search->{'vars'}->{'latitude'}->{'val'},$search->{'vars'}->{'longitude'}->{'val'});

    debug "remote search: $lat and $lon";

    # normalize the lat/lon to decimals
    my $position = new LLI::Position ( { 'lat' => $lat, 'lon' => $lon});
    $position->units_latlon("decimal");

    # should/could check for radius units here, but for now we just know about the cell
    #  a cell in this case is .001 x .001 lat/lon bounding box surrounding the lat,lon centroid
    my ($radius,$radius_units) = split(/\s/,$search->{'vars'}->{'radius'}->{'val'});
    $radius = int($radius) * .001;  #umm, and this isn't really a "radius", because we work in rectangles

    debug "radius: $radius";

    my $remote_components = $route->get_routes_search($position,$radius);
    debug ("found how many remote components in our radius/search criteria?",$remote_components);

    return 1 unless $remote_components; # no remote components, but that's ok.

    foreach my $component (@{ $remote_components })
      {
	# relay search/return node
	$component = @{$component}[0] if (ref $component eq "ARRAY");

	if (lc $component eq lc $cfg->{"component"}->{"publichost"})
	  {
	    # oop, this is a reflexive search, don't want that
	    debug "oop, we don't remote search ourselves";
	    next;
	  }
	debug "doing remote search on THIS $component";

	# send a search packet to the component
	my ($response_xml) = request_search($component,$search,$searchid);
	next unless ($response_xml);

	# get back the node that results will be published to
	my ($results) = parse_search_results($response_xml,$searchid);

	return 0 unless $results;

	debug ("parsed into ",$results);

	# and store it in our reference table
	my ($referenceid) = $store->save_remote_references($results);

	debug "results saved as referenceid: $referenceid"; 
      }

    return 1;
  }

sub request_search
  {
    my ($component,$search,$searchid) = @_;

    # this relays a request to a remote component -- almost a remote procedure call

    # <iq type='set' from='bob@jabber.org' id="data3">
    #  <query xmlns='jabber:iq:search'>
    #   <x xmlns='jabber:x:data' type='submit'>
    #     <field var='gender'><value>male</value></field>
    #     <field var='first'><value>Bob</value></field>
    #   </x>
    #  </query>
    #  </iq>

    my (@fields);

    foreach my $var (keys %{$search->{'vars'}})
      {
	push (@fields, { 'var' => $var,
			 'value' => [ $search->{'vars'}->{$var}->{'val'} ]
		       });
      }

    my $iq = {
	      'iq'  => { 'type' => 'set',
			 'from' => $cfg->{"component"}->{"searcher"}->{"publichost"},
			 'origfrom' => $cfg->{"component"}->{"publichost"},
			 'to'   => $component,
			 'id'   => $searchid,
			 'query'=> {
				    'xmlns' => "jabber:iq:search",
				    'x' =>
				    {
				     'xmlns' => 'jabber:x:data',
				     'type'  => 'submit',
				     'field' => [ @fields ]
				    }
				   }
		       }
             };

    my ($xml) = create_xml($iq);
    
    debug "sending out: $xml";
    
    my $nf = new Jabber::NodeFactory(fromstr => 1);
    my $node = $nf->newNodeFromStr($xml);

    my ($response) = $c->ask($node);
    return unless $response;
    $response = $response->toStr();

    debug ("send got back",$response);
    return $response;
}


sub parse_search_results
  {    
    # now that we've requested our search, store the association between
    # our searchID and the returned search node/component for later reference

    my ($response,$searchid) = @_;

    my (%results);
      
  READ_FORM:
    {
      my $twig = XML::Twig->new;
      debug "twig created";
      next unless $twig->safe_parse($response); #wraps in an eval, returns 0 on failure
      # should probably error check better here!

      debug "parsed ok";

      my $root = $twig->root;

    EXTRACT_RESULTS:
      {
	my (@search_array) = ("node","expiration");
	
	my @xs = $root->get_xpath("/iq/query/x");

	foreach my $x (@xs)
	  {
	    my $ns = lc $x->att('xmlns');

	    debug ("ns: $ns",$x->sprint());
	    
	    #next unless $ns eq 'jabber:iq:search';
	    #next unless $ns eq 'jabber:x:data';
	    
	    my ($items) = $x->first_child("item");

	    my (@fields) = $items->children("field");
	    
	    debug ("fields found",scalar(@fields));

	    foreach my $field (@fields)
	      {
		# finally, we're inside where the data is...
		foreach my $key (@search_array)
		  {
		    debug "looking for $key";
		    if ($field->att('var') eq $key)
		      {
			$results{$key} = $field->first_child("value");
			$results{$key} = $results{$key}->text if $results{$key};
			debug "found ".$results{$key};
		      }
		  }
	      }
	  }
	
	$results{'jid'} = $root->att('origfrom') || $root->att('from');
	$results{'searchid'} = $searchid;
      }
    }
    
    return (\%results);
  }

sub create_xml
  {
    my ($object) = @_;
    
    my ($xml) = eval
      {
	my $xs = new XML::Simple (forcearray => 1, keeproot => 1, keyattr => []);
	return $xs->XMLout($object);
      };
    
    # Get rid of new lines (this is legal, right?)
    $xml =~ s/\>\n+\</\>\</g;
    $xml =~ s/\>\s+\</\>\</g;
    
    debug $xml;
    
    return $xml;
  }

sub delete_search
  {
    my ($search_id) = @_;

    $store->delete_search($search_id);
    debug "$search_id removed from database";
  }

sub is_valid_search
  {
    my ($search) = @_;

    # $thing = {
    #          'node_name' => 'generic/HJYUNIRYZQDWLNOADVBNVPEKZSTCOH',
    #          'expiration' => '20030324121859',
    #          'owner' => 'mankins@jabber.media.mit.edu',
    #          'vars' => {
    #                      'latitude' => {
    #                                      'var' => 'latitude',
    #                                      'val' => '23.2333'
    #                                    },
    #                      'expiration' => {
    #                                        'var' => 'expiration',
    #                                        'val' => '86400'
    #                                      },
    #                      'longitude' => {
    #                                       'var' => 'longitude',
    #                                       'val' => '-123.002'
    #                                     },
    #                      'radius' => {
    #                                       'var' => 'radius',
    #                                       'val' => '77ft'
    #                                     }
    #                    }
    #        };

    # to be a valid search, we need at least a lat,lon .
    #  default radius = 1 minute
    #  default expiration = 60 seconds

    if (defined $search->{'vars'}->{'latitude'} && defined $search->{'vars'}->{'longitude'})
      {
	# now that we know we have a lat/lon, check for radius, defaulting as needed

	unless (defined $search->{'vars'}->{'radius'}->{'val'})
	  {
	    $search->{'vars'}->{'radius'}->{'val'} = "1 cell";
	  }	

	return (1,$search);
      }
    else
      {
	# oop, not a valid search

	debug "not a valid search";
	return (0,$search);
      }
  }

sub handle_iq
  {
    debug ("IQ:",@_);
    #exit;
  }

sub cleanup
  {
    debug "exiting";
    exit;
  }

sub debug
  {
    return unless $cfg->{'debug'}->{'level'};
    
    my ($time,@debug_data);

  DO_TIME:
    {
      my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
      
      my $month =   ("Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec") [(localtime)[4]];
      $year += 1900;
      
      if ($min < 10)
        {
          $time = $month ." ". $mday . ", ". $year. " ". $hour. ":0". $min. ":". $sec;
        }
      else
        {
          $time = $month ." ". $mday . ", ". $year. " ". $hour. ":". $min. ":". $sec;
        }
      
    }
    
    my ($line, $subname) = (caller(1))[2,3];
    
    my (@debug_data_buffer) = @_;
    
    foreach my $thing (@debug_data_buffer)
      {
        if (ref $thing)
          {
            $thing = "\n". Dumper $thing;
          }
	
        push(@debug_data,$thing);
      }
    
    my $hostname = color("blue"). $cfg->{'component'}->{'searcher'}->{'publichost'} . color("reset");
    $time = color("red reverse") . $time . color("reset");
    $line = color("green reverse") . $line . color("reset");
    $subname = color("yellow reverse") . $subname . color("reset");
    
    print DEBUG join(' : ', $hostname, $time, "line $line", "in sub $subname\n", color("reverse cyan"), @debug_data, color("reset")."\n");
  }

1;
